﻿using System;
using System.Runtime.InteropServices;
using Windows.Graphics.Imaging;
using Windows.Storage.Streams;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;

namespace TwisterCSharpUnsafe
{
    [Guid("905a0fef-bc53-11df-8c49-001e4fc686da"),
        InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    interface IBufferByteAccess
    {
        unsafe void Buffer(out byte* pByte);
    }

    public sealed partial class MainPage : Page
    {
        static readonly long cycleDuration = TimeSpan.FromSeconds(3).Ticks;

        // Bitmap information
        WriteableBitmap bitmap;
        byte[] srcPixels;
        unsafe byte* pDstPixels;
        int width, height, xCenter, yCenter;

        TimeSpan previousTimeSpan;

        public MainPage()
        {
            this.InitializeComponent();
            Application.Current.DebugSettings.EnableFrameRateCounter = true;
            Loaded += OnMainPageLoaded;
        }

        async void OnMainPageLoaded(object sender, RoutedEventArgs args)
        {
            Uri uri = new Uri("http://www.charlespetzold.com/pw6/PetzoldJersey.jpg");
            RandomAccessStreamReference streamRef = RandomAccessStreamReference.CreateFromUri(uri);

            using (IRandomAccessStreamWithContentType fileStream = await streamRef.OpenReadAsync())
            {
                BitmapDecoder decoder = await BitmapDecoder.CreateAsync(fileStream);
                BitmapFrame frame = await decoder.GetFrameAsync(0);

                // I know the parameterless version of GetPixelDataAsync works for this image
                PixelDataProvider pixelProvider = await frame.GetPixelDataAsync();
                srcPixels = pixelProvider.DetachPixelData();

                // Create the WriteableBitmap
                bitmap = new WriteableBitmap((int)frame.PixelWidth, (int)frame.PixelHeight);
            }

            // Set the bitmap to the Image element
            image.Source = bitmap;

            // Save some information in fields
            width = bitmap.PixelWidth;
            height = bitmap.PixelHeight;
            xCenter = width / 2;
            yCenter = height / 2;

            // Get a pointer to the bitmap pixels
            IBuffer iBuffer = bitmap.PixelBuffer;
            IBufferByteAccess bufferByteAccess = (IBufferByteAccess)iBuffer;
            unsafe
            {
                bufferByteAccess.Buffer(out pDstPixels);
            }

            // Let's go!
            CompositionTarget.Rendering += OnCompositionTargetRendering;
        }

        void OnCompositionTargetRendering(object sender, object args)
        {
            // Get elapsed time
            TimeSpan timeSpan = (args as RenderingEventArgs).RenderingTime;

            // Calculate twistAngle from -180 to 180 degrees
            double t = (timeSpan.Ticks % cycleDuration) / (double)cycleDuration;
            double tprime = 2 * (t < 0.5 ? t : 1 - t);
            double twistAngle = 2 * (tprime - 0.5) * Math.PI;

            for (int yDst = 0; yDst < height; yDst++)
                for (int xDst = 0; xDst < width; xDst++)
                {
                    // Calculate length of point to center and angle
                    int xDelta = xDst - xCenter;
                    int yDelta = yDst - yCenter;
                    double distanceToCenter = Math.Sqrt(xDelta * xDelta +
                                                        yDelta * yDelta);
                    double angleClockwise = Math.Atan2(yDelta, xDelta);

                    // Calculation angle of rotation for twisting effect
                    double xEllipse = xCenter * Math.Cos(angleClockwise);
                    double yEllipse = yCenter * Math.Sin(angleClockwise);
                    double radius = Math.Sqrt(xEllipse * xEllipse +
                                              yEllipse * yEllipse);
                    double fraction = Math.Max(0, 1 - distanceToCenter / radius);
                    double twist = fraction * twistAngle;

                    // Calculate the source pixel for each destination pixel
                    int xSrc = (int)(xCenter + (xDst - xCenter) * Math.Cos(twist)
                                             - (yDst - yCenter) * Math.Sin(twist));
                    int ySrc = (int)(yCenter + (xDst - xCenter) * Math.Sin(twist)
                                             + (yDst - yCenter) * Math.Cos(twist));
                    xSrc = Math.Max(0, Math.Min(width - 1, xSrc));
                    ySrc = Math.Max(0, Math.Min(height - 1, ySrc));

                    // Calculate the indices
                    int iDst = 4 * (yDst * width + xDst);
                    int iSrc = 4 * (ySrc * width + xSrc);

                    // Transfer the pixel bytes
                    unsafe
                    {
                        pDstPixels[iDst++] = srcPixels[iSrc++];
                        pDstPixels[iDst++] = srcPixels[iSrc++];
                        pDstPixels[iDst++] = srcPixels[iSrc++];
                        pDstPixels[iDst] = srcPixels[iSrc];
                    }
                }

            // Invalidate the bitmap
            bitmap.Invalidate();
        }
    }
}
